# This shell script emits a C file. -*- C -*-
#   Copyright 2004 Free Software Foundation, Inc.
#   Copyright 2004 Martin Kgler
#   Contributed by Martin Kgler (mkoegler@auto.tuwien.ac.at)
#
# This file is part of GLD, the Gnu Linker.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street - Fifth Floor, Boston, MA 02110-1301, USA.
#
cat >>e${EMULATION_NAME}.c <<EOF

void ld_find_biggest_input_section (lang_statement_list_type * list,
				    bfd_vma * maxsize,
				    lang_statement_union_type ** maxsection,
				    lang_statement_list_type ** maxlist);

void ld_move_input_section (lang_output_section_statement_type * to,
			    lang_statement_list_type * from,
			    lang_statement_union_type * element);

void ld_move_sections (const char *from, const char *to,
		       const char *aktsize, const char *maxsize);


void
ld_find_biggest_input_section (lang_statement_list_type * list,
			       bfd_vma * maxsize,
			       lang_statement_union_type ** maxsection,
			       lang_statement_list_type ** maxlist)
{
  lang_statement_union_type *p = list->head;
  while (p)
    {
      switch (p->header.type)
	{
	case lang_input_section_enum:
	  if (bfd_get_section_size (p->input_section.section) > *maxsize)
	    {
	      *maxsize = bfd_get_section_size (p->input_section.section);
	      *maxlist = list;
	      *maxsection = p;
	    }
	  break;
	case lang_wild_statement_enum:
	  ld_find_biggest_input_section (&p->wild_statement.children,
					 maxsize, maxsection, maxlist);
	  break;
	case lang_assignment_statement_enum:
	  break;
	default:
	  FAIL ();
	}
      p = p->header.next;
    }
}

void
ld_move_input_section (lang_output_section_statement_type * to,
		       lang_statement_list_type * from,
		       lang_statement_union_type * element)
{
  lang_statement_union_type **q = &from->head, *p = from->head;

  int found = 0;
  while (p)
    {
      if (p == element)
	{
	  found = 1;
	  break;
	}
      q = &p->header.next;
      p = p->header.next;
    }
  if (!found)
    FAIL ();

  *q = p->header.next;
  if (from->tail == &p->header.next)
    from->tail = q;

  element->input_section.section->output_section = 0;
  lang_add_section (&to->children, element->input_section.section, to);
}

void
ld_move_sections (const char *from, const char *to,
		  const char *aktsize, const char *maxsize)
{
  etree_type *akt, *max;
  bfd_vma aktval, maxval;
  int again = 1;

  akt = exp_nameop (NAME, aktsize);
  max = exp_nameop (NAME, maxsize);

  lang_size_sections (NULL, FALSE);

  while (again)
    {
      again = 0;
      lang_do_assignments ();
      lang_reset_memory_regions ();
      lang_size_sections (NULL, FALSE);

      aktval = exp_get_abs_int (akt, 0, (char*)aktsize);
      maxval = exp_get_abs_int (max, 0, (char*)maxsize);

      if (aktval > maxval)
	{
	  lang_output_section_statement_type *fromsec, *tosec;
	  lang_statement_union_type *element;
	  lang_statement_list_type *list;
	  bfd_vma maxsize = 0;

	  fromsec = lang_output_section_find (from);
	  tosec = lang_output_section_find (to);

	  if (!from)
	    {
	      einfo (_("Section '%s' not found; skipping section movement\n"),
		     from);
	      lang_reset_memory_regions ();
	      return;
	    }

	  if (!to)
	    {
	      einfo (_("Section '%s' not found; skipping section movement\n"),
		     to);
	      lang_reset_memory_regions ();
	      return;
	    }
	  ld_find_biggest_input_section (&fromsec->children,
					 &maxsize, &element, &list);

	  if (maxsize == 0)
	    {
	      einfo (_
		     ("Nothing found to move from '%s'; skipping section movement\n"),
		     from);
	      lang_reset_memory_regions ();
	      return;
	    }

	  ld_move_input_section (tosec, list, element);
	  again = 1;
	}

    }
  lang_reset_memory_regions ();
}

/* Flag for the emulation-specific "--move-sections" option.  */
static bfd_boolean move_sections = FALSE;
static name_list *section_move_list = 0;
static name_list **section_move_tail = &section_move_list;

static void m68hc05_move_sections (void);
static void append (name_list *** list, char *name);
static void m68hc05_syslib (char *name);

static void
append (name_list *** list, char *name)
{
  name_list *element = (name_list *) xmalloc (sizeof (name_list));

  element->name = name;
  element->next = (name_list *) NULL;
  **list = element;
  *list = &element->next;
}

static void
m68hc05_move_sections ()
{
  name_list *p = section_move_list;
  while (p)
    {
      char *from, *to;
      char *aktsize, *maxsize;
      char *ptr = xstrdup (p->name);
      from = ptr;
      while (*ptr && *ptr != ':')
	ptr++;
      if (*ptr != ':')
	{
	  einfo (_("wrong section move command: '%s'\n"), p->name);
	  p = p->next;
	  continue;
	}
      *ptr = 0;
      ptr++;

      to = ptr;
      while (*ptr && *ptr != ':')
	ptr++;
      if (*ptr != ':')
	{
	  einfo (_("wrong section move command: '%s'\n"), p->name);
	  p = p->next;
	  continue;
	}
      *ptr = 0;
      ptr++;

      aktsize = ptr;
      while (*ptr && *ptr != ':')
	ptr++;
      if (*ptr != ':')
	{
	  einfo (_("wrong section move command: '%s'\n"), p->name);
	  p = p->next;
	  continue;
	}
      *ptr = 0;
      ptr++;

      maxsize = ptr;

      ld_move_sections (from, to, aktsize, maxsize);

      p = p->next;
    }
}

/* To not introduce a change in LD itself, SYSLIB is used to indicate section moves */
static void
m68hc05_syslib (char *name)
{
  append (&section_move_tail, name);
}

static void
elf_m68hc05_before_allocation (void)
{
  if(move_sections)
    m68hc05_move_sections();
  gld${EMULATION_NAME}_before_allocation ();
}

EOF

# Define some shell vars to insert bits of code into the standard elf
# parse_args and list_options functions.
#
PARSE_AND_LIST_PROLOGUE='
#define OPTION_MOVE_SECTIONS                 301
'

PARSE_AND_LIST_LONGOPTS='
  { "move-sections", no_argument, NULL, OPTION_MOVE_SECTIONS},
'

PARSE_AND_LIST_OPTIONS='
  fprintf (file, _("  --move-sections\t\tenables move of input sections to other other output sections, if an output section becomes to big\n"));
'

PARSE_AND_LIST_ARGS_CASES='
    case OPTION_MOVE_SECTIONS:
      move_sections = TRUE;
      break;
'

LDEMUL_BEFORE_ALLOCATION=elf_m68hc05_before_allocation 
LDEMUL_SYSLIB=m68hc05_syslib
